package org.bitbucket.bradleysmithllc.etlunit;

import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestClass;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestMethod;

import java.util.ArrayList;
import java.util.List;

public interface StatusReporter {
	enum test_completion_code {
		success,
		failure,
		warning,
		error
	}

	class CompletionStatus {
		private List<TestExecutionError> testExecutionErrors = new ArrayList<TestExecutionError>();

		private final List<TestAssertionFailure> assertionFailures = new ArrayList<TestAssertionFailure>();
		private final List<String> assertionFailureIds = new ArrayList<String>();
		private final List<TestWarning> testWarnings = new ArrayList<TestWarning>();
		private final List<String> testWarningIds = new ArrayList<String>();

		private String expectedErrorId;

		private List<String> expectedFailureIds = new ArrayList<String>();
		private List<String> expectedWarningIds = new ArrayList<String>();

		private boolean expectedErrorOverride = false;
		private boolean expectedFailureOverride = false;

		private boolean resolved = false;

		public void setExpectedErrorOverride(boolean expectedErrorOverride) {
			this.expectedErrorOverride = expectedErrorOverride;
		}

		public void setExpectedFailureOverride(boolean expectedFailureOverride) {
			this.expectedFailureOverride = expectedFailureOverride;
		}

		public String getExpectedErrorId() {
			return expectedErrorId;
		}

		public void setExpectedErrorId(String expectedErrorId) {
			this.expectedErrorId = expectedErrorId;
		}

		public void addExpectedFailureId(String expectedErrorId) {
			expectedFailureIds.add(expectedErrorId);
		}

		public void addExpectedFailureIds(List<String> expectedErrorIds) {
			expectedFailureIds.addAll(expectedErrorIds);
		}

		public void addExpectedWarningId(String expectedErrorId) {
			expectedWarningIds.add(expectedErrorId);
		}

		public void addExpectedWarningIds(List<String> expectedErrorIds) {
			expectedWarningIds.addAll(expectedErrorIds);
		}

		public List<TestExecutionError> getErrors() {
			return testExecutionErrors;
		}

		public boolean hasError() {
			return testExecutionErrors.size() != 0;
		}

		public boolean hasFailures() {
			return hasError() || assertionFailures.size() != 0;
		}

		public List<TestAssertionFailure> getAssertionFailures() {
			return assertionFailures;
		}

		public List<TestWarning> getWarnings() {
			return testWarnings;
		}

		public void addWarning(TestWarning warn) {
			String warningId = warn.getWarningId();
			if (!testWarningIds.contains(warningId)) {
				testWarningIds.add(warningId);
			}

			testWarnings.add(warn);
		}

		public void addFailure(TestAssertionFailure failure) {
			String failureId = failure.getFailureId();
			if (!assertionFailureIds.contains(failureId)) {
				assertionFailureIds.add(failureId);
			}

			assertionFailures.add(failure);
		}

		public void addError(TestExecutionError err) {
			testExecutionErrors.add(err);
		}

		public String toString() {
			return "errors=" + testExecutionErrors + ", failures=" + assertionFailures + ", warnings" + testWarnings;
		}

		public test_completion_code getTestResult() {
			resolveStatus();

			// check for any status involving an error
			if (testExecutionErrors.size() != 0) {
				// check for expected error override
				if (!expectedErrorOverride) {
					// stop - this is an error
					return test_completion_code.error;
				}
			}

			if (assertionFailureIds.size() != 0) {
				// might be a failure
				if (!expectedFailureOverride) {
					return test_completion_code.failure;
				}
			}

			return test_completion_code.success;

			/*
			&& expectedErrorId == null)
			{
				testResult = test_completion_code.error;
			}
			else if (assertionFailures.size() != 0)
			{
				testResult = test_completion_code.failure;
			}
			else if (testWarnings.size() != 0)
			{
				testResult = test_completion_code.warning;
			}
			else
			{
				testResult = test_completion_code.success;
			}

			switch (testResult)
			{
				case success:
					return test_completion_code.success;
				case warning:
					return test_completion_code.success;
				case failure:
				case error:
					if (expectedErrorOverride)
					{
						// success *iif* there were no failures
						return assertionFailureIds.size() != 0 ? test_completion_code.failure : test_completion_code.success;
					}

					if (expectedFailureOverride)
					{
						// success *iif* there were no failures
						return test_completion_code.success;
					}

					return testResult;
				default:
					throw new UnsupportedOperationException();
			}
			 */
		}

		public List<String> getExpectedWarningIds() {
			return expectedWarningIds;
		}

		public List<String> getExpectedFailureIds() {
			return expectedFailureIds;
		}

		public List<String> getTestWarningIds() {
			return testWarningIds;
		}

		public List<String> getAssertionFailureIds() {
			return assertionFailureIds;
		}

		public void resolveStatus() {
			if (resolved) {
				return;
			}

			resolved = true;

			List<String> failureIds = getExpectedFailureIds();

			if (failureIds.size() != 0) {
				List<String> assertionFailureIds = new ArrayList<String>(getAssertionFailureIds());
				if (failureIds.containsAll(assertionFailureIds) && assertionFailureIds.containsAll(failureIds)) {
					// this test passes
					setExpectedFailureOverride(true);
				} else {
					addFailure(new TestAssertionFailure("Failure(s) "
							+ failureIds
							+ " expected but "
							+ assertionFailureIds
							+ " caught", TestConstants.FAIL_EXPECTED_ASSERTION));
				}
			}

			String expectedErrorId = getExpectedErrorId();

			if (expectedErrorId != null) {
				if (hasError()) {
					if (getErrors().size() != 0 && !getErrors().get(0).getErrorId().equals(expectedErrorId)) {
						// mark the error as okay so the assertion failure will win
						setExpectedErrorOverride(true);
						// make sure this is cancelled
						setExpectedFailureOverride(false);
						addFailure(new TestAssertionFailure("Error '"
								+ expectedErrorId
								+ "' expected but '"
								+ getErrors()
								+ "' caught", TestConstants.FAIL_EXPECTED_EXCEPTION));
					} else {
						// this test passes
						setExpectedErrorOverride(true);
					}
				} else {
					addFailure(new TestAssertionFailure("Error '" + expectedErrorId + "' expected",
							TestConstants.FAIL_EXPECTED_EXCEPTION));
				}
			}

			List<String> warningIds = getExpectedWarningIds();

			if (warningIds.size() != 0) {
				List<String> testWarningIds = getTestWarningIds();
				if (warningIds.containsAll(testWarningIds) && testWarningIds.containsAll(warningIds)) {
					// nothing to do here because warnings can't fail a method except for expected warnings not happening
				} else {
					addFailure(new TestAssertionFailure("Warning(s) "
							+ warningIds
							+ " expected but "
							+ testWarningIds
							+ " caught", TestConstants.FAIL_EXPECTED_WARNING));
				}
			}
		}
	}

	void scanStarted();

	void scanCompleted();

	void testsStarted(int numTestsSelected);

	void testClassAccepted(ETLTestClass method);

	void testMethodAccepted(ETLTestMethod method);

	void testBeginning(ETLTestMethod method);

	void testCompleted(ETLTestMethod method, CompletionStatus status);

	void testsCompleted();
}